// -*- C++ -*-
/*!
 * @file  MonteCarloLocalizationLRF.cpp
 * @brief Monte Carlo Localization using LRF
 * @date $Date$
 *
 * $Id$
 */

#include "MonteCarloLocalizationLRF.h"
#include "mcl.h"
#include "map.h"
#include "const.h"

// GTK関連
#ifndef _WITHOUT_GUI
 #include "interface.h"
 #include "support.h"
 #include "draw.h"
#endif

// Module specification
// <rtc-template block="module_spec">
static const char* montecarlolocalizationlrf_spec[] =
  {
    "implementation_id", "MonteCarloLocalizationLRF",
    "type_name",         "MonteCarloLocalizationLRF",
    "description",       "Monte Carlo Localization using LRF",
    "version",           "0.8.1",
    "vendor",            "Tsuyoshi Suenaga, NAIST",
    "category",          "localization",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default._num_particles", "300",
    "conf.default._def_pos_x_m", "7.0",
    "conf.default._def_pos_y_m", "6.5",
    "conf.default._def_pos_theta_deg", "180.0",
    "conf.default._std_pos_x_m", "0.5",
    "conf.default._std_pos_y_m", "0.5",
    "conf.default._std_pos_theta_deg", "10.0",
    "conf.default.__m_per_pixel", "0.1",
    "conf.default._screen_offset_x", "200",
    "conf.default._screen_offset_y", "200",
    "conf.default._map_format", "OCCUPANCY_GRID_MAP",
    "conf.default._map_filename", "mapdata/NAIST1F-occupancyGridMap.pgm",
    "conf.default._x_coord_front_in_range_data", "0",
    "conf.default._x_coord_front_in_current_pose", "1",
    "conf.default._x_coord_front_out_current_pose", "1",
    "conf.default._x_coord_front_in_odometry", "1",

    ""
  };
// </rtc-template>

MonteCarloLocalizationLRF::MonteCarloLocalizationLRF(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_RangeDataIn("RangeData", m_RangeData),
    m_OdometryIn("Odometry", m_Odometry),
    m_CurrentPosInitIn("CurrentPosInit", m_CurrentPosInit),
    m_CurrentPosOut("CurrentPos", m_CurrentPos),
    m_LocationInfoPort("LocationInfo"),
    m_LRFInfoPort("LRFInfo"),
    m_MapInfoPort("MapInfo"),
    // </rtc-template>
    m_Location(mcl, &pose_mean, &pose_std),
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("RangeData", m_RangeDataIn);
  registerInPort("Odometry", m_OdometryIn);
  registerInPort("CurrentPosInit", m_CurrentPosInitIn);
  
  // Set OutPort buffer
  registerOutPort("CurrentPos", m_CurrentPosOut);
  
  // Set service provider to Ports
  m_LocationInfoPort.registerProvider("Location", "LocationInfo", m_Location);
  
  // Set service consumers to Ports
  m_LRFInfoPort.registerConsumer("LRF", "LRFInfo", m_LRF);
  m_MapInfoPort.registerConsumer("Map", "MapInfo", m_Map);
  
  // Set CORBA Service Ports
  registerPort(m_LRFInfoPort);
  registerPort(m_MapInfoPort);
  registerPort(m_LocationInfoPort);
  
  // </rtc-template>

// GTK関連
#ifndef _WITHOUT_GUI
  initialize_for_gtk();
#endif

}

MonteCarloLocalizationLRF::~MonteCarloLocalizationLRF()
{
}



RTC::ReturnCode_t MonteCarloLocalizationLRF::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("_num_particles", m__num_particles, "300");
  bindParameter("_def_pos_x_m", m__def_pos_x_m, "7.0");
  bindParameter("_def_pos_y_m", m__def_pos_y_m, "6.5");
  bindParameter("_def_pos_theta_deg", m__def_pos_theta_deg, "180.0");
  bindParameter("_std_pos_x_m", m__std_pos_x_m, "0.5");
  bindParameter("_std_pos_y_m", m__std_pos_y_m, "0.5");
  bindParameter("_std_pos_theta_deg", m__std_pos_theta_deg, "10.0");
  bindParameter("_m_per_pixel", m__m_per_pixel, "0.1");
  bindParameter("_screen_offset_x", m__screen_offset_x, "200");
  bindParameter("_screen_offset_y", m__screen_offset_y, "200");
  bindParameter("_map_format", m__map_format, "OCCUPANCY_GRID_MAP");
  bindParameter("_map_filename", m__map_filename, "mapdata/NAIST1F-occupancyGridMap.pgm");
  bindParameter("_x_coord_front_in_range_data", m__x_coord_front_lrf_info, "0");
  bindParameter("_x_coord_front_in_current_pose", m__x_coord_front_current_pos_init, "1");
  bindParameter("_x_coord_front_out_current_pose", m__x_coord_front_current_pos, "1");
  bindParameter("_x_coord_front_in_odometry", m__x_coord_front_odometry, "1");
  
  // </rtc-template>

  mcl = new Mcl;
  mcl->filter = NULL;

  return RTC::RTC_OK;
}


RTC::ReturnCode_t MonteCarloLocalizationLRF::onFinalize()
{
  delete mcl;
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t MonteCarloLocalizationLRF::onActivated(RTC::UniqueId ec_id)
{
  PortServiceList* portlist;

  portlist = this->get_ports();
  port_connect_check = new bool [portlist->length()];
  for(unsigned int i=0; i<portlist->length(); i++){
    PortService_ptr port;
    port = (*portlist)[i];
    if(port->get_port_profile()->connector_profiles.length()!=0){
      port_connect_check[i]=true;
    }else{
      port_connect_check[i]=false;
    }
  }

  std::cerr << "LRF Information : front coordinate is ";
  if(m__x_coord_front_lrf_info){
    std::cerr << "X.";
  }else{
    std::cerr << "Y.";
  }
  std::cerr << std::endl;

  std::cerr << "Input Current Pose : front coordinate is ";
  if(m__x_coord_front_current_pos_init){
    std::cerr << "X.";
  }else{
    std::cerr << "Y.";
  }

  std::cerr << "Output Current Pose : front coordinate is ";
  if(m__x_coord_front_current_pos){
    std::cerr << "X.";
  }else{
    std::cerr << "Y.";
  }

  std::cerr << "Input Odometry : front coordinate is ";
  if(m__x_coord_front_odometry){
    std::cerr << "X.";
  }else{
    std::cerr << "Y.";
  }

// GTK関連
#ifndef _WITHOUT_GUI
  // ウィンドウの作成
  //window1 = create_window1 ();
  gui = create_window1 ();
  // gtkスタート
  g_thread_create((GThreadFunc)gtk_main_thread, NULL, FALSE, NULL);
  initialize_for_gui(gui);

  // 表示のオフセットを設定する
  gui->screen_offset_x = m__screen_offset_x;
  gui->screen_offset_y = m__screen_offset_y;
  // 表示の解像度を設定する
  gui->screen_resolution = m__m_per_pixel; // [m/pixel]
#endif


  // サービスポートからLRF情報の取得
  if(port_connect_check[4]==true){
    mcl->lrf.total_points = m_LRF->getTotalScanPoints();
    mcl->lrf.resolution = m_LRF->getResolution();
    if(m__x_coord_front_lrf_info){
      // X軸正面 -> Y軸正面
      mcl->lrf.offset_fov = m_LRF->getOffsetFOV()+M_PI/2.0;
    }else{
      // Y軸正面
      mcl->lrf.offset_fov = m_LRF->getOffsetFOV();
    }
    
    mcl->lrf.fov = (mcl->lrf.total_points - 1) * mcl->lrf.resolution;
    mcl->lrf.range.resize(mcl->lrf.total_points);
    
    std::cerr << "TotalPoints: " << mcl->lrf.total_points << std::endl;
    std::cerr << "Resolution : " << mcl->lrf.resolution << std::endl;
    std::cerr << "OffsetFOV  : " << mcl->lrf.offset_fov << std::endl;
    std::cerr << "FieldOfView: " << mcl->lrf.fov << std::endl;
  }

  // パーティクル作成
  mcl->filter = localizer_particle_filter_new(m__num_particles);
  mcl->prev_filter = localizer_particle_filter_new(m__num_particles);

  // 初期位置，姿勢の平均，標準偏差を設定
  pose_mean.x = m__def_pos_x_m; // [m]
  pose_mean.y = m__def_pos_y_m; // [m]
  pose_mean.theta = DEG2RAD(m__def_pos_theta_deg); // [rad]
  pose_std.x = m__std_pos_x_m; // [m]
  pose_std.y = m__std_pos_y_m; // [m]
  pose_std.theta = DEG2RAD(m__std_pos_theta_deg); // [rad]
  
  // パーティクル初期化
  localizer_initialize_particles_gaussians(mcl->filter, 1, &pose_mean, &pose_std);

  // オドメトリの初期化
  mcl->odometry.x = 0;
  mcl->odometry.y = 0;
  mcl->odometry.theta = 0;
  mcl->prev_odometry.x = 0;
  mcl->prev_odometry.y = 0;
  mcl->prev_odometry.theta = 0;

  if(port_connect_check[5]==true){
    RTC::TimedOctetSeq mapdata;
    CORBA::Short x, y, width, height;
    CORBA::Double resolution;
    if(m_Map->getFullGridMap(mapdata, x, y, width, height, resolution)!=false){
      mcl->ogmap.origin_x = x;
      mcl->ogmap.origin_y = y;
      mcl->ogmap.width  = width;
      mcl->ogmap.height = height;
      mcl->ogmap.resolution = resolution;
      for(int i=0;i<width*height;i++)
      mcl->ogmap.data[i] = mapdata.data[i];
      std::cout << "Received Grid Map: " << std::endl;
      std::cout << "  (x,y)=(" << x << "," << y << ")" << std::endl;
      std::cout << "  (w,h)=(" << width << "," << height << ")" << std::endl;
      std::cout << "  resolution=" << resolution << std::endl;
    }else{
      std::cout << "Could not receive grid map." << std::endl;
      return RTC::RTC_ERROR;
    }
    convert_occupancy_gridmap_to_pointmap(&mcl->ogmap, &mcl->pmap, 200);
  }else{
    // マップ読み込み
    if(!m__map_format.find("POINT_MAP")){
      // ポイントマップ
      //load_point_map("mapdata/pointmap_5f_cm.dat", &mcl->pmap, 0.01); // [cm] -> [m]
      load_point_map(m__map_filename.c_str(), &mcl->pmap, 0.01); // [cm] -> [m]
    }else if(!m__map_format.find("LINE_MAP")){
      // ラインマップ
      //load_line_map("mapdata/linemap_5f_cm.dat", &mcl->lmap, 0.01); // [cm] -> [m]
      load_line_map(m__map_filename.c_str(), &mcl->lmap, 0.01); // [cm] -> [m]
      convert_linemap_to_pointmap(&mcl->lmap ,&mcl->pmap, 0.07); // [m/grid]
    }else if(!m__map_format.find("OCCUPANCY_GRID_MAP")){
      // 占有グリッドマップ
      //load_occupancy_grid_map("mapdata/NAIST1F-occupancyGridMap.pgm", &mcl->ogmap);
      load_occupancy_grid_map(m__map_filename.c_str(), &mcl->ogmap);
      convert_occupancy_gridmap_to_pointmap(&mcl->ogmap, &mcl->pmap, 200);
    }else{
      return RTC::RTC_ERROR;
    }
  }

  return RTC::RTC_OK;
}



RTC::ReturnCode_t MonteCarloLocalizationLRF::onDeactivated(RTC::UniqueId ec_id)
{
  localizer_particle_filter_delete(mcl->filter);
  localizer_particle_filter_delete(mcl->prev_filter);

  delete port_connect_check;

// GTK関連
#ifndef _WITHOUT_GUI
  // ウィンドウ終了
  destroy_window(gui);
  // gtk終了
  gtk_main_quit();
#endif
  return RTC::RTC_OK;
}



RTC::ReturnCode_t MonteCarloLocalizationLRF::onExecute(RTC::UniqueId ec_id)
{
  int i;
  int width, height;
  double likelihood;

  if(m_OdometryIn.isNew()){
    // オドメトリの取得
    m_OdometryIn.read();
    if(m__x_coord_front_odometry){
      // X軸正面 -> Y軸正面
      mcl->odometry.x = -m_Odometry.y;
      mcl->odometry.y = m_Odometry.x;
    }else{
      // Y軸正面
      mcl->odometry.x = m_Odometry.x;
      mcl->odometry.y = m_Odometry.y;
    }
    mcl->odometry.theta = m_Odometry.theta;
  }

  if(m_CurrentPosInitIn.isNew()){
    // 取得位置を初期位置，姿勢として初期化
    m_CurrentPosInitIn.read();
    pose_mean.x = m_CurrentPosInit.x; // [m]
    pose_mean.y = m_CurrentPosInit.y; // [m]
    if(m__x_coord_front_current_pos_init){
      // X軸正面 -> Y軸正面
      pose_mean.theta = m_CurrentPosInit.theta-M_PI/2.0; // [rad]
    }else{
      // Y軸正面
      pose_mean.theta = m_CurrentPosInit.theta; // [rad]
    }
    pose_std.x = m__std_pos_x_m; // [m]
    pose_std.y = m__std_pos_y_m; // [m]
    pose_std.theta = DEG2RAD(m__std_pos_theta_deg); // [rad]
    
    // パーティクル初期化
    localizer_initialize_particles_gaussians(mcl->filter, 1, &pose_mean, &pose_std);
  }

  if(m_RangeDataIn.isNew()){
    // データの取得
    m_RangeDataIn.read();
    for(i=0; i<mcl->lrf.total_points; i++){
      mcl->lrf.range.at(i) = m_RangeData.data[i]/1000.0; // mm -> m
    }
    //std::cout << mcl->lrf.range[0] << std::endl;

    // MCL
    if(port_connect_check[1]==false){
      // 等分散モデルによるサンプリング
      localizer_particle_resample(mcl->filter);
    }else{
      // オドメトリ動作モデルによるサンプリング
      sample_motion_model_odometry(mcl->filter, &mcl->odometry, &mcl->prev_odometry);
    }
    mcl->filter->sum_weight = 0;
    mcl->filter->max_particle.weight = 0;
    for( i=0 ; i<mcl->filter->num_particles; i++){
      if(port_connect_check[5]==true || !m__map_format.find("OCCUPANCY_GRID_MAP")){
	// 尤度
	likelihood = likelihood_field_range_finder_model( &mcl->lrf,
							  mcl->filter->particles[i],
							  &mcl->ogmap);
      }else{
	likelihood = likelihood_field_range_finder_model( &mcl->lrf,
							  mcl->filter->particles[i],
							  &mcl->pmap);
      }

      mcl->filter->sum_weight+= likelihood;
      mcl->filter->particles[i].weight = likelihood;
      if ( likelihood > mcl->filter->max_particle.weight ){
	mcl->filter->max_particle = mcl->filter->particles[i];
      }
    }
    // normalize particle
    for( i = 0 ; i < mcl->filter->num_particles; i++){
      mcl->filter->particles[i].weight /= mcl->filter->sum_weight;
    }
    mcl->filter->sum_weight = 1.0;
    mcl->filter->ready= 1;
    memcpy(mcl->prev_filter->particles, mcl->filter->particles, 
	   sizeof(LocalizerParticle)*mcl->filter->num_particles);
    std::cout << "(x,y,theta) = (" \
	      << mcl->filter->max_particle.x << ", " \
	      << mcl->filter->max_particle.y << ", " \
	      << RAD2DEG(mcl->filter->max_particle.theta) << ")  " \
	      << mcl->filter->max_particle.weight << std::endl;

    // 現在位置データの送信
    m_CurrentPos.x = mcl->filter->max_particle.x;
    m_CurrentPos.y = mcl->filter->max_particle.y;
    if(m__x_coord_front_current_pos){
      // Y軸正面 -> X軸正面
      m_CurrentPos.theta = mcl->filter->max_particle.theta+M_PI/2.0;
    }else{
      // Y軸正面
      m_CurrentPos.theta = mcl->filter->max_particle.theta;
    }
    m_CurrentPosOut.write();
  }

// GTK関連
#ifndef _WITHOUT_GUI
  gdk_threads_enter();
  
  // pixmapを初期化する
  draw_area_initialize(gui);
  
  // pixmapにマップを描画
  draw_pointmap(&mcl->pmap, gui);
  
  if(mcl->filter != NULL) {
    // パーティクルの描画
    draw_particles( gui,
		    mcl->filter->particles, 
		    mcl->filter->num_particles,
		    10);
    // pixmapにセンサ中心を描画
    //draw_point(gui, &pose_mean, 4, make_color(0x9999, 0x9999, 0x9999));
    draw_particle(gui, &mcl->filter->max_particle, 10,
		  make_color(0xfb00,0xe700,0x1800),
		  make_color(0xff00,0x0,0x0));
    // LRFデータをpixmapに描画
    draw_laser_range(gui, &mcl->lrf, &mcl->filter->max_particle);
  }
  // pixmapを描画
  show_draw_area(gui);

  // ウィンドウサイズを取得する
  get_draw_size(gui, &width, &height);
//   fprintf(stderr, "(%4d,%4d)\n",
// 	  width, height);
  gdk_threads_leave();
#endif
  
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t MonteCarloLocalizationLRF::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void MonteCarloLocalizationLRFInit(RTC::Manager* manager)
  {
    RTC::Properties profile(montecarlolocalizationlrf_spec);
    manager->registerFactory(profile,
                             RTC::Create<MonteCarloLocalizationLRF>,
                             RTC::Delete<MonteCarloLocalizationLRF>);
  }
  
};


